using System.Net.Mime;
using System.Xml;
using Autofac;
using FitConnect.Encryption;
using FitConnect.Exceptions;
using FitConnect.Interfaces;
using FitConnect.Interfaces.Subscriber;
using FitConnect.Models;
using FitConnect.Models.v1.Api;
using FitConnect.Services.Models;
using FitConnect.Services.Models.v1.Destination;
using FitConnect.Services.Models.v1.Submission;
using Microsoft.Extensions.Logging;
using Microsoft.IdentityModel.Tokens;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;
using Newtonsoft.Json.Schema;
using JsonSchema = NJsonSchema.JsonSchema;
using ValidationError = NJsonSchema.Validation.ValidationError;

namespace FitConnect;

/// <summary>
///     Fluent API for the FitConnect.Subscriber
/// </summary>
public class Subscriber : FitConnectClient, IReceivedSubmission,
    ISubscriber {
    private IContainer? _container;

    /// <summary>
    /// Can be used in TEST Environment to skip the data validation
    /// </summary>
    public bool SkipDataValidation { get; set; }


    internal Subscriber(IContainer container) : base(FitConnectEnvironment.Test,
        container.Resolve<IFitConnectSettings>().SubscriberClientId,
        container.Resolve<IFitConnectSettings>().SubscriberClientSecret,
        container,
        container.Resolve<IFitConnectSettings>().PrivateDecryptionKeys,
        container.Resolve<IFitConnectSettings>().PrivateSigningKey) {
        Encryption = new FitEncryption(
            container.Resolve<IFitConnectSettings>().PrivateDecryptionKeys,
            container.Resolve<IFitConnectSettings>().PrivateSigningKey,
            container.Resolve<ILogger>());
        _container = container;
    }

    public Subscriber(FitConnectEnvironment environment,
        string clientId, string clientSecret,
        List<string> privateDecryptionKeys,
        string privateKeySigning,
        ILogger? logger = null) : base(environment, clientId, clientSecret, logger,
        privateDecryptionKeys, privateKeySigning) {
        Encryption = new FitEncryption(privateDecryptionKeys, privateKeySigning, logger);
    }


    /// <summary>
    ///     Receives the available submissions from the server
    /// </summary>
    /// <param name="destinationId"></param>
    /// <param name="skip"></param>
    /// <param name="take"></param>
    /// <returns></returns>
    public async Task<IEnumerable<SubmissionForPickup>> GetAvailableSubmissionsAsync(
        Guid destinationId,
        int skip = 0,
        int take = 100) {
        try {
            var submissionsResult =
                await SubmissionService.ListSubmissionsAsync(destinationId, 0, 100);

            // Creating a dictionary of destinationId to submissionIds from the REST API result
            return submissionsResult.Submissions ?? new List<SubmissionForPickup>();
        }
        catch (Exception e) {
            Logger?.LogError(e, "Error while receiving submissions");
            throw new FitConnectSubscriberException("Error while receiving submissions", e);
        }
    }

    public async Task<IReceivedSubmission>
        RequestSubmissionAsync(SubmissionForPickup submission) =>
        await RequestSubmissionAsync(submission.Id);

    /// <summary>
    ///     Receives a specific submission from the server and verifies submission<br />
    ///     https://docs.fitko.de/fit-connect/docs/receiving/verification/
    /// </summary>
    /// <param name="submissionId"></param>
    /// <returns></returns>
    public async Task<IReceivedSubmission> RequestSubmissionAsync(Guid submissionId) {
        try {
            var submission = (Submission)await SubmissionService.GetSubmissionAsync(submissionId);
            return await CheckSubmissionAsync(submission);
        }
        catch (Exception e) {
            Logger?.LogError(e, "Error while receiving submission");
            throw new FitConnectSubscriberException("Error while receiving submission", e);
        }
    }

    internal async Task<IReceivedSubmission> CheckSubmissionAsync(Submission submission) {
        var (submitEvent, metadataSignature) = await CheckSecurityEventTokensAsync(submission);


        if (metadataSignature != submission.EncryptedMetadata
                ?.Split(ProjectSpecification.TokenSeparator).Last()) {
            var problem = new Problems(Problems.ProblemTypeEnum.IncorrectAuthenticationTag,
                Problems.TitleAuthenticationTagInvalid,
                Problems.DetailAuthenticationMetadataInvalid,
                Problems.ProblemInstanceEnum.Metadata);
            await RejectSubmissionAsync(submission, problem);
            throw new SecurityEventException(problem);
        }

        string? metadataString;
        try {
            var (metadataStringResult, _, metaHash) =
                Encryption.Decrypt(submission.EncryptedMetadata!);
            metadataString = metadataStringResult;

            if (metadataString == null)
                throw new Exception("Unable to decrypt metadata");
        }
        catch (Exception e) {
            var problem = new Problems(Problems.ProblemTypeEnum.EncryptionIssue,
                Problems.DetailDecryptionIssueMetadata);
            await RejectSubmissionAsync(submission, problem);
            throw new SecurityEventException(problem, e);
        }


        submission.Metadata = JsonConvert.DeserializeObject<Metadata>(metadataString);
        if (submission.Metadata?.SchemaUri == null) {
            var problem = new Problems(Problems.ProblemTypeEnum.MissingSchema,
                Problems.TitleMissingSchema,
                Problems.DetailMissingSchema,
                Problems.ProblemInstanceEnum.Metadata);
            // To activate in a later version, when SchemaUri is required
            // await RejectSubmissionAsync(submission, problem);
            // throw new SecurityEventException(problem);
        }
        else {
            JsonSchema? schema;
            try {
                schema = await JsonSchema.FromUrlAsync(submission.Metadata?.SchemaUri.ToString());
            }
            catch (Exception e) {
                var problem = new Problems(Problems.ProblemTypeEnum.UnsupportedMetaSchema,
                    Problems.DetailUnsupportedSchema);
                await RejectSubmissionAsync(submission, problem);
                throw new SecurityEventException(problem, e);
            }

            await VerifyMetadata(submission, metadataString, schema);
        }


        if (submission.Metadata == null) {
            var problem = new Problems(Problems.ProblemTypeEnum.SyntaxViolation,
                Problems.DetailSyntaxViolationMetadata);
            await RejectSubmissionAsync(submission, problem);
            throw new SecurityEventException(problem);
        }

        var destination = await DestinationService.GetDestinationAsync(submission.DestinationId);
        if (destination == null) throw new ArgumentException("Destination not found");

        var destinationService = destination.Services!.FirstOrDefault(s =>
            s.Identifier == submission.ServiceType.Identifier);
        if (destinationService == null) {
            var problem = new Problems(Problems.ProblemTypeEnum.UnsupportedService,
                Problems.DetailUnsupportedService);
            await RejectSubmissionAsync(submission, problem);
            throw new SecurityEventException(problem);
        }

        await CheckReplyChannelAsync(submission, destination);

        string? dataString = null;
        if (submission.EncryptedData != null)
            try {
                (dataString, _, _) =
                    Encryption.Decrypt(submission.EncryptedData);
                submission.Data = dataString;

                if (dataString == null)
                    throw new ArgumentException(
                        "DataString can not be null if decryption succeeded");
            }
            catch (Exception e) {
                var problem = new Problems(Problems.ProblemTypeEnum.EncryptionIssue,
                    Problems.DetailDecryptionIssueMetadata,
                    Problems.DetailDecryptionIssueData,
                    Problems.ProblemInstanceEnum.Data);
                await RejectSubmissionAsync(submission, problem);
                throw new SecurityEventException(problem, e);
            }

        await VerifyDataHashAsync(submission, dataString!);

        if (submission.Data == null) {
            var problem = new Problems(Problems.ProblemTypeEnum.MissingData,
                Problems.DetailMissingData);
            await RejectSubmissionAsync(submission, problem);
            throw new SecurityEventException(problem);
        }


        if (submission.DataMimeType == MediaTypeNames.Application.Json) {
            var dataSchema = GetSchemaUrlFromJson(submission.Data);
            if (dataSchema == null)
                Logger?.LogWarning(
                    "$schema attribute missing in submission data of submission {SubmissionId}",
                    submission.Id);
        }


        if (destinationService.SubmissionSchemas != null && !destinationService
                .SubmissionSchemas
                .Select(s => s.SchemaUri.ToString())
                .Contains(submission.Metadata.ContentStructure.Data.SubmissionSchema.SchemaUri
                    .ToString())) {
            var problem = new Problems(Problems.ProblemTypeEnum.UnsupportedDataSchema,
                Problems.DetailUnsupportedDataSchema);
            await RejectSubmissionAsync(submission, problem);
            throw new SecurityEventException(problem, null,
                $"\r\n\tExpected: {destinationService.SubmissionSchemas.Select(s => s.SchemaUri.ToString()).Aggregate((a, b) => a + ", " + b)}" +
                $"\r\n\tActual: {submission.Metadata.ContentStructure.Data.SubmissionSchema.SchemaUri}");
        }


        if (submission.DataMimeType == MediaTypeNames.Application.Json)
            if (JsonConvert.DeserializeObject(submission.Data) == null) {
                var problem = new Problems(
                    Problems.ProblemTypeEnum.SyntaxViolation,
                    Problems.TitleSyntaxViolation,
                    Problems.DetailSyntaxViolationDataJson,
                    Problems.ProblemInstanceEnum.Data);
                await RejectSubmissionAsync(submission, problem);
                throw new SecurityEventException(problem);
            }

        if (submission.DataMimeType == MediaTypeNames.Application.Xml)
            try {
                var doc = new XmlDocument();
                doc.LoadXml(submission.Data);
            }
            catch (Exception e) {
                var problem = new Problems(
                    Problems.ProblemTypeEnum.SyntaxViolation,
                    Problems.TitleSyntaxViolation,
                    Problems.DetailSyntaxViolationDataXml,
                    Problems.ProblemInstanceEnum.Data);
                await RejectSubmissionAsync(submission, problem);
                throw new SecurityEventException(problem, e);
            }

        submission.Attachments = await DownloadAttachmentsAsync(submission);

        await CheckAttachments(submission, submitEvent);

        Submission = submission;
        return this;
    }

    private async Task RejectSubmissionAsync(ISubmissionBody submission, Problems[] problems) {
        await CompleteSubmission(submission, FinishSubmissionStatus.Rejected,
            problems);
    }

    public async Task RejectSubmissionAsync(ISubmissionBody submission,
        Problems problem, params Problems[] problems) {
        var rejectionProblems = new List<Problems> { problem };

        rejectionProblems.AddRange(problems);
        await RejectSubmissionAsync(submission, rejectionProblems.ToArray());
    }

    public async Task<SubmissionStatus> GetStatusForSubmission(Guid destinationId, Guid caseId,
        Guid submissionId) {
        try {
            return await base.GetStatusForSubmissionAsync(
                new SentSubmission(submissionId, caseId, destinationId, null), false);
        }
        catch (Exception e) {
            Logger?.LogError("Error getting status for submission");
            throw new FitConnectSubscriberException("Error getting status for submission", e);
        }
    }

    public Submission? Submission { get; private set; }


    public Uri? DataSchemaUri =>
        Submission?.Metadata?.ContentStructure.Data.SubmissionSchema.SchemaUri;

    public Guid? DestinationId => Submission?.DestinationId;

    /// <summary>
    ///     Reading attachments for a submission.
    /// </summary>
    /// <returns></returns>
    /// <summary>
    ///     Reading attachments for a submission.
    /// </summary>
    /// <returns></returns>
    public async Task<IEnumerable<Attachment>> GetAttachmentsAsync() {
        try {
            if (Submission?.Id == null || Submission?.Metadata == null)
                throw new Exception("No submission available");

            var attachments = new List<Attachment>();
            foreach (var id in Submission!.AttachmentIds) {
                var encryptedAttachment =
                    await SubmissionService.GetAttachmentAsync(Submission.Id, id);
                var (_, content, hash) = Encryption.Decrypt(encryptedAttachment);
                var attachmentMeta =
                    Submission.Metadata.ContentStructure.Attachments.First(
                        a => a.AttachmentId == id);

                var item = Attachment.FromReceivedSubmission(id, attachmentMeta, content,
                    encryptedAttachment.Split(ProjectSpecification.TokenSeparator).Last());
                item.Description = attachmentMeta.Description;
                item.Purpose = attachmentMeta.Purpose;
                item.MimeType = attachmentMeta.MimeType;
                attachments.Add(item);
            }

            Submission.Attachments = attachments;
            return attachments;
        }
        catch (Exception e) {
            throw new FitConnectSubscriberException(
                "Could not get attachments for submission", e);
        }
    }

    public async Task AcceptSubmissionAsync(params Problems[] problems) {
        await CompleteSubmissionAsync(Submission!, FinishSubmissionStatus.Accepted, problems);
    }

    public async Task RejectSubmissionAsync(params Problems[] problems) {
        await CompleteSubmissionAsync(Submission!, FinishSubmissionStatus.Rejected, problems);
    }

    public string GetDataMimeType() =>
        Submission!.Metadata!.ContentStructure.Data.SubmissionSchema.MimeType ??
        throw new FitConnectSubscriberException("Submission was not loaded correctly");

    public Uri? GetDataSchemaUri() =>
        Submission?.Metadata?.ContentStructure.Data.SubmissionSchema.SchemaUri;


    private string? GetSchemaUrlFromJson(string? jsonString) {
        return jsonString == null
            ? null
            : (JsonConvert.DeserializeObject(jsonString) as JObject)?["$schema"]?.ToString();
    }

    /// <summary>
    ///     Check Hash of attachment
    /// </summary>
    /// <param name="submission"></param>
    /// <returns></returns>
    /// <exception cref="SecurityEventException"></exception>
    private async Task<List<Attachment>> DownloadAttachmentsAsync(Submission submission) {
        var attachments = new List<Attachment>();
        foreach (var id in submission!.AttachmentIds) {
            string encryptedAttachment;
            try {
                encryptedAttachment = await SubmissionService.GetAttachmentAsync(submission.Id, id);
            }
            catch (Exception e) {
                var problem = new Problems(Problems.ProblemTypeEnum.MissingAttachments,
                    Problems.TitleAttachmentMissing,
                    string.Format(Problems.DetailAttachmentMissing, id),
                    $"attachment:{id}");
                await RejectSubmissionAsync(submission, problem);
                throw new SecurityEventException(problem, e);
            }

            byte[]? content;
            byte[]? hash;
            try {
                (_, content, hash) = Encryption.Decrypt(encryptedAttachment);
            }
            catch (Exception e) {
                var problem = new Problems(
                    Problems.ProblemTypeEnum.EncryptionIssue,
                    Problems.TitleDecryptionIssue,
                    string.Format(Problems.DetailDecryptionIssueAttachment, id),
                    $"attachment:{id}"
                );
                await RejectSubmissionAsync(submission, problem);
                throw new SecurityEventException(problem, e);
            }

            var attachmentMeta =
                submission.Metadata?.ContentStructure.Attachments.First(a =>
                    a.AttachmentId == id);

            if (attachmentMeta != null) {
                attachments.Add(Attachment.FromReceivedSubmission(id, attachmentMeta, content,
                    encryptedAttachment.Split(ProjectSpecification.TokenSeparator).Last()));


                if (attachmentMeta?.Hash.Content != FitEncryption.CalculateHash(content).Content) {
                    Logger?.LogWarning(
                        "Wrong hash for attachment {attachment}\nHash should be {should} but is {is}",
                        id, FitEncryption.CalculateHash(content).Content,
                        attachmentMeta?.Hash.Content);
                    var problem = new Problems(
                        Problems.ProblemTypeEnum.HashMismatch,
                        Problems.TitleHashMismatch,
                        string.Format(Problems.DetailHashMismatchAttachment, id),
                        $"attachment:{id}");
                    await RejectSubmissionAsync(submission, problem);
                    throw new SecurityEventException(problem);
                }
            }
        }

        return attachments;
    }

    private async Task CompleteSubmission(ISubmissionBody submission,
        FinishSubmissionStatus status, Problems[]? problems = null) {
        if (status != FinishSubmissionStatus.Rejected && problems != null)
            throw new ArgumentException("Problems can only be set for rejected submissions");

        string token;
        switch (status) {
            case FinishSubmissionStatus.Rejected:
                token = await Encryption.CreateRejectSecurityEventToken(submission.SubmissionId,
                    submission.CaseId, submission.DestinationId,
                    problems ??
                    throw new ArgumentException("On reject problems must be provided"));
                break;
            case FinishSubmissionStatus.Accepted:
                token = await Encryption.CreateAcceptSecurityEventToken(submission);
                break;
            default:
                throw new ArgumentOutOfRangeException(nameof(status), status, null);
        }

        Logger?.LogDebug("Token to accept submission: {Token}", token);
        var result = await CasesService.FinishSubmissionAsync(submission.CaseId, token);
        Logger?.LogInformation("Submission completed {Status}", result);
    }


    /// <summary>
    ///     Accept or reject submission
    /// </summary>
    private async Task CompleteSubmissionAsync(ISubmissionBody submission,
        FinishSubmissionStatus status, Problems[]? problems = null) {
        try {
            string? token;
            switch (status) {
                case FinishSubmissionStatus.Rejected:
                    token = await Encryption.CreateRejectSecurityEventToken(submission.SubmissionId,
                        submission.CaseId, submission.DestinationId,
                        problems ??
                        throw new ArgumentException("On reject problems must be provided"));
                    break;
                case FinishSubmissionStatus.Accepted:
                    if (submission is Submission sub) {
                        var authenticationTags =
                            FitEncryption.GenerateAuthenticationTags(sub);
                        token = await Encryption.CreateAcceptSecurityEventToken(submission,
                            problems,
                            authenticationTags);
                    }
                    else {
                        throw new ArgumentException(
                            "Only full qualified submission allowed to accept");
                    }

                    break;
                default:
                    throw new ArgumentOutOfRangeException(nameof(status), status, null);
            }

            Logger?.LogDebug("Token to accept submission: {Token}", token);
            var result = await CasesService.FinishSubmissionAsync(submission.CaseId, token);
            Logger?.LogInformation("Submission completed {Status}", result);
        }
        catch (Exception e) {
            Logger?.LogError(e, "Error while completing submission");
            throw new FitConnectSubscriberException("Error while completing submission", e);
        }
    }


    private enum FinishSubmissionStatus {
        Accepted,
        Rejected
    }

    #region Check Submission and Reject if needed

    protected override async Task<ICollection<ValidationError>> CheckDataSchemaAsync(
        Submission submission) {
        var isValidSchemaData = await base.CheckDataSchemaAsync(submission);

        if (!isValidSchemaData.IsNullOrEmpty()) {
            var problem = new Problems(Problems.ProblemTypeEnum.SchemaViolation,
                Problems.DetailsSchemaViolationData);
            await RejectSubmissionAsync(submission, problem);
            throw new SecurityEventException(problem);
        }

        return isValidSchemaData;
    }


    private async Task VerifyDataHashAsync(Submission submission, string dataString) {
        if (submission.Metadata?.ContentStructure.Data.Hash.Content ==
            FitEncryption.CalculateHash(dataString).Content) return;

        Logger?.LogWarning("Data hash mismatch: {DataHash} != {CalculatedHash}",
            submission.Metadata?.ContentStructure.Data.Hash.Content,
            FitEncryption.CalculateHash(dataString));


        var problem = new Problems(Problems.ProblemTypeEnum.HashMismatch,
            Problems.TitleHashMismatch,
            Problems.DetailHashMismatchData, Problems.ProblemInstanceEnum.Data);
        await RejectSubmissionAsync(submission, problem);
        throw new SecurityEventException(problem);
    }


    /// <summary>
    ///     Checking metadata
    /// </summary>
    /// <param name="submission"></param>
    /// <param name="metadataString"></param>
    /// <param name="schema"></param>
    /// <exception cref="Exception"></exception>
    private async Task VerifyMetadata(Submission submission, string metadataString,
        JsonSchema schema) {
        var valid = await JsonHelper.VerifyMetadata(metadataString, schema);
        if (!valid) {
            Logger?.LogWarning("Invalid metadata: {Metadata}", metadataString);


            var problem = new Problems(Problems.ProblemTypeEnum.UnsupportedMetaSchema,
                Problems.DetailUnsupportedSchema);
            await RejectSubmissionAsync(submission, problem);
            throw new SecurityEventException(problem);
        }
    }


    /// <summary>
    ///     Compares the reply channel of the submission with the allowed channels of the destination
    /// </summary>
    /// <param name="submission"></param>
    /// <param name="destination"></param>
    /// <exception cref="SecurityEventException"></exception>
    private async Task CheckReplyChannelAsync(Submission submission,
        PublicDestinationDto destination) {
        var rcSubmission = submission.Metadata?.ReplyChannel;
        var rcDestination = destination.ReplyChannels;

        if (rcSubmission == null) return;

        if (rcSubmission.Elster != null && rcDestination?.Elster != null) return;
        if (rcSubmission.Fink != null && rcDestination?.Fink != null) return;
        if (rcSubmission.DeMail != null && rcDestination?.DeMail != null) return;
        if (rcSubmission.EMail != null && rcDestination?.EMail != null) return;

        var problems = new Problems(Problems.ProblemTypeEnum.UnsupportedReplyChannel,
            Problems.DetailUnsupportedReplyChannel);
        await RejectSubmissionAsync(submission, problems);
        throw new SecurityEventException(problems);
    }

    /// <summary>
    ///     Checking:<br />
    ///     - Exactly one submit element<br />
    ///     - Authentication Tag of submit event<br />
    ///     - Authentication Tag of submission data
    /// </summary>
    /// <param name="submission"></param>
    /// <returns></returns>
    /// <exception cref="SecurityEventException"></exception>
    private async Task<(SecurityEventToken, string)> CheckSecurityEventTokensAsync(
        Submission submission) {
        List<SecurityEventToken> status;
        try {
            status = await GetEventLogAsync(submission);
        }
        catch (Exception? e) {
            Logger?.LogWarning("Could not get status for submission: {SubmissionId}",
                submission.Id);
            await RejectSubmissionAsync(submission,
                new Problems(Problems.ProblemTypeEnum.InvalidEventLog,
                    Problems.DetailEventLogInconsistent));

            throw new SecurityEventException(Problems.TitleEventLogInconsistent,
                Problems.DetailEventLogInconsistent, e);
        }

        if (status.Where(s => s.SubmissionId == submission.Id)
                .Count(set => set.EventType == EventType.Submit) != 1) {
            await RejectSubmissionAsync(submission,
                new Problems(Problems.ProblemTypeEnum.InvalidEventLog,
                    Problems.DetailEventLogNotExactlyOneSubmit));
            throw new SecurityEventException(Problems.TitleEventLogInconsistent,
                Problems.DetailEventLogNotExactlyOneSubmit);
        }

        var submitEvent = status.First(set => set.EventType == EventType.Submit);
        var authenticationTag = (submitEvent.Events?.SubmitSubmissionEvent?.AuthenticationTags);

        var dataSignature = authenticationTag?.Data;
        var metadataSignature = authenticationTag?.Metadata ?? "";

        if (submission.EncryptedData?.Split(ProjectSpecification.TokenSeparator).Last() !=
            dataSignature) {
            var problem = new Problems(Problems.ProblemTypeEnum.IncorrectAuthenticationTag,
                Problems.DetailHashMismatchData);
            await RejectSubmissionAsync(submission, problem);
            throw new SecurityEventException(problem);
        }

        if (submission.EncryptedMetadata?.Split(ProjectSpecification.TokenSeparator).Last() !=
            metadataSignature) {
            var problem = new Problems(Problems.ProblemTypeEnum.IncorrectAuthenticationTag,
                Problems.DetailAuthenticationMetadataInvalid);
            await RejectSubmissionAsync(submission, problem);
            throw new SecurityEventException(problem);
        }

        return (submitEvent, metadataSignature);
    }

    /// <summary>
    ///     Checking Attachments<br />
    ///     - Verify list of attachments<br />
    ///     - Authentication tags of attachments
    /// </summary>
    /// <exception cref="SecurityEventException"></exception>
    private async Task CheckAttachments(Submission submission,
        SecurityEventToken submitEvent) {
        if (submission?.Attachments != null) {
            // @formatter:off
            
            if (submission.Attachments.Count != submission.Metadata?.ContentStructure.Attachments.Count 
                || submission.Attachments.Count != (submitEvent.Events?.SubmitSubmissionEvent?.AuthenticationTags?.Attachments?.Count ?? 0)
                || !submission.Attachments.TrueForAll(a =>submission.Metadata?.ContentStructure.Attachments?.Any(ma=>ma.AttachmentId == a.Id) ?? false)
                || !submission.Attachments.TrueForAll(a =>submitEvent.Events?.SubmitSubmissionEvent?.AuthenticationTags?.Attachments?.ContainsKey(a.Id) ?? false)
                ) {
                var problem = new Problems(Problems.ProblemTypeEnum.AttachmentMismatch,
                    Problems.DetailAttachmentsMismatch);
                await RejectSubmissionAsync(submission, problem);
                throw new SecurityEventException(problem);
            }
            // @formatter:on

            var attachmentAuthenticationTags =
                submitEvent.Events!.SubmitSubmissionEvent!.AuthenticationTags!.Attachments;

            var problems = new List<Problems>();
            foreach (var attachment in submission.Attachments)
                if (attachmentAuthenticationTags?[attachment.Id] !=
                    attachment.AttachmentAuthentication) {
                    var problem = new Problems(Problems.ProblemTypeEnum.IncorrectAuthenticationTag,
                        string.Format(Problems.DetailAuthenticationAttachmentInvalid,
                            attachment.Id));
                    problems.Add(problem);
                }

            if (problems.Any()) {
                await RejectSubmissionAsync(submission, problems.ToArray());
                throw new SecurityEventException(problems.ToArray());
            }
        }
    }

    #endregion
}
